# stdlib
import contextlib
import socket
from typing import List
from typing import Tuple
from typing import Union

# third party
import docker
from result import Result

# relative
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...store.document_store import SyftSuccess
from ...types.uid import UID
from ...util.telemetry import instrument
from ..service import AbstractService
from ..service import AuthedServiceContext
from ..service import SyftError
from ..service import service_method
from ..user.user_roles import ADMIN_ROLE_LEVEL
from .utils import DEFAULT_WORKER_POOL_NAME
from .utils import _get_healthcheck_based_on_status
from .worker_pool import ContainerSpawnStatus
from .worker_pool import SyftWorker
from .worker_pool import WorkerHealth
from .worker_pool import WorkerStatus
from .worker_pool import _get_worker_container_status
from .worker_stash import WorkerStash

WORKER_NUM = 0


def get_main_backend():
    hostname = socket.gethostname()
    return f"{hostname}-backend-1"


def start_worker_container(
    worker_num: int, context: AuthedServiceContext, syft_worker_uid
):
    client = docker.from_env()
    existing_container_name = get_main_backend()
    hostname = socket.gethostname()
    worker_name = f"{hostname}-worker-{worker_num}"
    return create_new_container_from_existing(
        worker_name=worker_name,
        client=client,
        existing_container_name=existing_container_name,
        syft_worker_uid=syft_worker_uid,
    )


def create_new_container_from_existing(
    worker_name: str,
    client: docker.client.DockerClient,
    existing_container_name: str,
    syft_worker_uid,
) -> docker.models.containers.Container:
    # Get the existing container
    existing_container = client.containers.get(existing_container_name)

    # Inspect the existing container
    details = existing_container.attrs

    # Extract relevant settings
    image = details["Config"]["Image"]
    command = details["Config"]["Cmd"]
    environment = details["Config"]["Env"]
    ports = details["NetworkSettings"]["Ports"]
    host_config = details["HostConfig"]

    volumes = {}
    for vol in host_config["Binds"]:
        parts = vol.split(":")
        key = parts[0]
        bind = parts[1]
        mode = parts[2]
        if "/storage" in bind:
            # we need this because otherwise we are using the same node private key
            # which will make account creation fail
            worker_postfix = worker_name.split("-", 1)[1]
            key = f"{key}-{worker_postfix}"
        volumes[key] = {"bind": bind, "mode": mode}

    # we need this because otherwise we are using the same node private key
    # which will make account creation fail

    environment = dict([e.split("=", 1) for e in environment])
    environment["CREATE_PRODUCER"] = "false"
    environment["N_CONSUMERS"] = 1
    environment["DOCKER_WORKER_NAME"] = worker_name
    environment["DEFAULT_ROOT_USERNAME"] = worker_name
    environment["DEFAULT_ROOT_EMAIL"] = f"{worker_name}@openmined.org"
    environment["PORT"] = str(8003 + WORKER_NUM)
    environment["HTTP_PORT"] = str(88 + WORKER_NUM)
    environment["HTTPS_PORT"] = str(446 + WORKER_NUM)
    environment["SYFT_WORKER_UID"] = str(syft_worker_uid)

    environment.pop("NODE_PRIVATE_KEY", None)

    new_container = client.containers.create(
        name=worker_name,
        image=image,
        command=command,
        environment=environment,
        ports=ports,
        detach=True,
        volumes=volumes,
        tty=True,
        stdin_open=True,
        network_mode=f"container:{existing_container.id}",
    )

    new_container.start()
    return new_container


@instrument
@serializable()
class WorkerService(AbstractService):
    store: DocumentStore
    stash: WorkerStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = WorkerStash(store=store)

    @service_method(
        path="worker.start_workers",
        name="start_workers",
        roles=ADMIN_ROLE_LEVEL,
    )
    def start_workers(
        self, context: AuthedServiceContext, n: int = 1
    ) -> Union[List[ContainerSpawnStatus], SyftError]:
        """Add a Container Image."""
        worker_pool_service = context.node.get_service("SyftWorkerPoolService")

        return worker_pool_service.add_workers(
            context, number=n, pool_name=DEFAULT_WORKER_POOL_NAME
        )

    @service_method(path="worker.get_all", name="get_all", roles=ADMIN_ROLE_LEVEL)
    def list(self, context: AuthedServiceContext) -> Union[SyftSuccess, SyftError]:
        """List all the workers."""
        result = self.stash.get_all(context.credentials)

        if result.is_err():
            return SyftError(message=f"Failed to fetch workers. {result.err()}")

        workers = result.ok()

        if context.node.in_memory_workers:
            return workers

        # If container workers, check their statuses
        for idx, worker in enumerate(workers):
            result = check_and_update_status_for_worker(
                worker=worker,
                worker_stash=self.stash,
                credentials=context.credentials,
            )
            if result.is_err():
                return SyftError(
                    message=f"Failed to update status for worker: {worker.id}. Error: {result.err()}"
                )
            workers[idx] = worker

        return workers

    @service_method(path="worker.status", name="status", roles=ADMIN_ROLE_LEVEL)
    def status(
        self,
        context: AuthedServiceContext,
        uid: UID,
    ) -> Union[Tuple[WorkerStatus, WorkerHealth], SyftError]:
        result = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if result.is_err():
            return SyftError(message=f"Failed to retrieve worker with UID {uid}")
        worker: SyftWorker = result.ok()

        if context.node.in_memory_workers:
            return worker.status, worker.healthcheck

        result = check_and_update_status_for_worker(
            worker=worker,
            worker_stash=self.stash,
            credentials=context.credentials,
        )

        if result.is_err():
            return SyftError(
                message=f"Failed to update status for worker: {worker.id}. Error: {result.err()}"
            )

        worker = result.ok()

        return worker.status, worker.healthcheck


def check_and_update_status_for_worker(
    worker: SyftWorker,
    worker_stash: WorkerStash,
    credentials: SyftVerifyKey,
) -> Result[SyftWorker, str]:
    with contextlib.closing(docker.from_env()) as client:
        worker_status = _get_worker_container_status(client, worker)

    if isinstance(worker_status, SyftError):
        return worker_status

    worker.status = worker_status

    worker.healthcheck = _get_healthcheck_based_on_status(status=worker_status)

    result = worker_stash.update(
        credentials=credentials,
        obj=worker,
    )

    return result



# stdlib
from typing import Optional

# relative
from ...custom_worker.config import WorkerConfig
from ...node.credentials import SyftVerifyKey
from ...serde.serializable import serializable
from ...types.datetime import DateTime
from ...types.syft_object import SYFT_OBJECT_VERSION_1
from ...types.syft_object import SyftObject
from ...types.uid import UID
from .image_identifier import SyftWorkerImageIdentifier


@serializable()
class SyftWorkerImage(SyftObject):
    __canonical_name__ = "SyftWorkerImage"
    __version__ = SYFT_OBJECT_VERSION_1

    __attr_unique__ = ["config"]
    __attr_searchable__ = ["config", "image_hash", "created_by"]
    __repr_attrs__ = ["image_identifier", "image_hash", "created_at", "built_at"]

    id: UID
    config: WorkerConfig
    image_identifier: Optional[SyftWorkerImageIdentifier]
    image_hash: Optional[str]
    created_at: DateTime = DateTime.now()
    created_by: SyftVerifyKey
    built_at: Optional[DateTime]



# stdlib
import contextlib
from typing import List
from typing import Optional
from typing import Tuple
from typing import Union

# third party
import docker
import pydantic

# relative
from ...custom_worker.config import DockerWorkerConfig
from ...serde.serializable import serializable
from ...store.document_store import DocumentStore
from ...types.datetime import DateTime
from ...types.dicttuple import DictTuple
from ...types.uid import UID
from ..context import AuthedServiceContext
from ..response import SyftError
from ..response import SyftSuccess
from ..service import AbstractService
from ..service import service_method
from ..user.user_roles import DATA_OWNER_ROLE_LEVEL
from .image_registry import SyftImageRegistry
from .image_registry_service import SyftImageRegistryService
from .utils import docker_build
from .utils import docker_push
from .worker_image import SyftWorkerImage
from .worker_image import SyftWorkerImageIdentifier
from .worker_image_stash import SyftWorkerImageStash


@serializable()
class SyftWorkerImageService(AbstractService):
    store: DocumentStore
    stash: SyftWorkerImageStash

    def __init__(self, store: DocumentStore) -> None:
        self.store = store
        self.stash = SyftWorkerImageStash(store=store)

    @service_method(
        path="worker_image.submit_dockerfile",
        name="submit_dockerfile",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def submit_dockerfile(
        self, context: AuthedServiceContext, docker_config: DockerWorkerConfig
    ) -> Union[SyftSuccess, SyftError]:
        image_identifier = SyftWorkerImageIdentifier(repo="", tag="")
        worker_image = SyftWorkerImage(
            config=docker_config,
            created_by=context.credentials,
            image_identifier=image_identifier,
        )
        res = self.stash.set(context.credentials, worker_image)

        if res.is_err():
            return SyftError(message=res.err())

        return SyftSuccess(
            message=f"Dockerfile <id: {worker_image.id}> successfully submitted."
        )

    @service_method(
        path="worker_image.build",
        name="build",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def build(
        self,
        context: AuthedServiceContext,
        image_uid: UID,
        tag: str,
        registry_uid: Optional[UID] = None,
    ) -> Union[SyftSuccess, SyftError]:
        registry: SyftImageRegistry = None

        result = self.stash.get_by_uid(credentials=context.credentials, uid=image_uid)
        if result.is_err():
            return SyftError(
                message=f"Failed to get image for uid: {image_uid}. Error: {result.err()}"
            )

        worker_image: SyftWorkerImage = result.ok()

        if registry_uid:
            # get registry from image registry service
            image_registry_service: SyftImageRegistryService = context.node.get_service(
                SyftImageRegistryService
            )
            result = image_registry_service.get_by_id(context, registry_uid)
            if result.is_err():
                return result
            registry: SyftImageRegistry = result.ok()

        try:
            if registry:
                image_identifier = SyftWorkerImageIdentifier.with_registry(
                    tag=tag, registry=registry
                )
            else:
                image_identifier = SyftWorkerImageIdentifier.from_str(tag=tag)
        except pydantic.ValidationError as e:
            return SyftError(message=f"Failed to create tag: {e}")

        # if image is already built and identifier is unchanged, return an error
        if (
            worker_image.built_at
            and worker_image.image_identifier
            and worker_image.image_identifier.full_name_with_tag
            == image_identifier.full_name_with_tag
        ):
            return SyftError(message=f"Image<{image_uid}> is already built")

        worker_image.image_identifier = image_identifier

        if not context.node.in_memory_workers:
            result = docker_build(worker_image)
            if isinstance(result, SyftError):
                return result

            worker_image.image_hash = result.image_hash
            worker_image.built_at = DateTime.now()

            result = SyftSuccess(
                message=f"Build {worker_image} succeeded.\n{result.logs}"
            )
        else:
            result = SyftSuccess(
                message="Image building skipped, since using InMemory workers."
            )

        update_result = self.stash.update(context.credentials, obj=worker_image)

        if update_result.is_err():
            return SyftError(
                message=f"Failed to update image meta information: {update_result.err()}"
            )

        return result

    @service_method(
        path="worker_image.push",
        name="push",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def push(
        self,
        context: AuthedServiceContext,
        image: UID,
        username: Optional[str] = None,
        password: Optional[str] = None,
    ) -> Union[SyftSuccess, SyftError]:
        if context.node.in_memory_workers:
            return SyftSuccess(
                message="Skipped pushing image, since using InMemory workers."
            )

        result = self.stash.get_by_uid(credentials=context.credentials, uid=image)
        if result.is_err():
            return SyftError(
                message=f"Failed to get image for uid: {image}. Error: {result.err()}"
            )
        worker_image: SyftWorkerImage = result.ok()

        if (
            worker_image.image_identifier is None
            or worker_image.image_identifier.registry_host == ""
        ):
            return SyftError(
                message=f"Image {worker_image} does not have a valid registry host."
            )
        elif worker_image.built_at is None:
            return SyftError(
                message=f"Image {worker_image} is not built yet. Please build it first."
            )

        result = docker_push(
            image=worker_image,
            username=username,
            password=password,
        )

        if isinstance(result, SyftError):
            return result

        return SyftSuccess(
            message=f'The image was successfully pushed to "{worker_image.image_identifier.full_name_with_tag}"'
        )

    @service_method(
        path="worker_image.get_all",
        name="get_all",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def get_all(
        self, context: AuthedServiceContext
    ) -> Union[DictTuple[str, SyftWorkerImage], SyftError]:
        """
        One image one docker file for now
        """
        result = self.stash.get_all(credentials=context.credentials)
        if result.is_err():
            return SyftError(message=f"{result.err()}")
        images: List[SyftWorkerImage] = result.ok()

        res: List[Tuple] = []
        for im in images:
            if im.image_identifier is not None:
                res.append((im.image_identifier.repo_with_tag, im))
            else:
                res.append(("default-worker-image", im))

        return DictTuple(res)

    @service_method(
        path="worker_image.delete",
        name="delete",
        roles=DATA_OWNER_ROLE_LEVEL,
    )
    def delete(
        self, context: AuthedServiceContext, uid: UID
    ) -> Union[SyftSuccess, SyftError]:
        #  Delete Docker image given image tag
        res = self.stash.get_by_uid(credentials=context.credentials, uid=uid)
        if res.is_err():
            return SyftError(message=f"{res.err()}")
        image: SyftWorkerImage = res.ok()

        if not context.node.in_memory_workers and image and image.image_identifier:
            try:
                full_tag: str = image.image_identifier.repo_with_tag
                with contextlib.closing(docker.from_env()) as client:
                    client.images.remove(image=full_tag)
            except docker.errors.ImageNotFound:
                return SyftError(message=f"Image {full_tag} not found.")
            except Exception as e:
                return SyftError(
                    message=f"Failed to delete image {full_tag}. Error: {e}"
                )

        result = self.stash.delete_by_uid(credentials=context.credentials, uid=uid)

        if result.is_err():
            return SyftError(message=f"{result.err()}")

        returned_message: str = (
            result.ok().message + f". Image {uid} deleted successfully."
        )

        return SyftSuccess(message=returned_message)
